==> code-fns-0.10.0.tgz: package/lib/esm/tags.js <==
import { createStarryNight, all } from '@wooorm/starry-night';
import style from './dark-style';
import wcwidth from 'wcwidth';
const starryNight = createStarryNight(all);
let starryNightCache = null;
export async function ready() {
  starryNightCache = await starryNight;
}
const handler = {
  get(_, language) {
    return (code, ...rest) => {
      return {
        language,
        spans: Array.from(code),
        nodes: rest
      };
    };
  }

};
export const language = new Proxy({}, handler);
export function parse(code, options) {
  const raw = integrate(reindent(code));
  if (starryNightCache == null) throw new Error(`you must await ready()`);
  const sn = starryNightCache;
  const scope = sn.flagToScope(code.language);

  if (typeof scope !== 'string') {
    throw new Error(`language ${code.language} not found`);
  }

  const parsed = sn.highlight(raw, scope);
  return parsed.children.map(child => colorRecurse(child, options?.codeStyle ?? {})).flat().map(({
    color,
    ...rest
  }) => ({ ...rest,
    color: color === '' ? '#c9d1d9' : color
  }));
}
const rules = new Map(Object.entries(style).map(([k, v]) => [k, new Map(Object.entries(v))]));
const styleMap = [[['pl-s'], s => s.stringContent?.text], [['pl-s', 'pl-pds'], s => s.stringPunctuation?.text], [['pl-c'], s => s.comment?.text], [['pl-smi'], s => s.variable?.text], [['pl-v'], s => s.parameter?.text], [['pl-s', 'pl-sr'], s => s.regexp?.content ?? s.regexp?.text], [['pl-s', 'pl-sr', 'pl-pds'], s => s.regexp?.brackets ?? s.regexp?.text], [['pl-s', 'pl-sr', 'pl-k'], s => s.regexp?.flags ?? s.regexp?.text], [['pl-c1'], s => s.literal?.text], [['pl-k'], s => s.keyword?.text], [['pl-en'], s => s.entityName?.text]];

function findAssociated(styleMap, target, notFound) {
  let best = notFound;
  let specificity = 0;

  for (const [keys, value] of styleMap) {
    if (keys.length <= specificity) {
      continue;
    }

    let keyIndex = 0;
    let targetIndex = 0;

    while (targetIndex < target.length && keyIndex < keys.length) {
      const key = keys[keyIndex];
      const targetKey = target[targetIndex];

      if (targetKey === key) {
        keyIndex++;
      }

      targetIndex++;
    }

    if (keyIndex === keys.length) {
      // all keys found
      best = value;
      specificity = keys.length;
    }
  }

  return best;
}

export function getColor(classList, codeStyle) {
  const fn = findAssociated(styleMap, classList, () => undefined);
  const result = fn(codeStyle);

  if (result != null) {
    return result;
  }

  const styles = rules.get(`.${classList.at(-1)}`);
  console.assert((styles?.size ?? 0) <= 1, `more styles than just color`);
  return styles?.get('color');
}

function colorRecurse(parsed, codeStyle, classList = []) {
  if (parsed.type === 'text') {
    return [{
      code: parsed.value,
      color: ''
    }];
  } else if (parsed.type === 'element') {
    const className = parsed.properties?.className ?? [];
    const color = getColor(classList.concat(className), codeStyle);
    const children = parsed.children.map(child => colorRecurse(child, codeStyle, classList.concat(className)));
    const result = [];

    const emit = () => {
      if (temp !== '') {
        if (color != '') {
          result.push({
            code: temp,
            color
          });
        } else {
          result.push({
            code: temp,
            color: ''
          });
        }

        temp = '';
      }
    };

    let temp = '';

    for (const child of children) {
      for (const item of child) {
        if (item.color === '') {
          temp += item.code;
        } else {
          emit();
          result.push(item);
        }
      }

      emit();
    }

    emit();
    return result;
  } else {
    throw new Error();
  }
}

export function toString(tokens) {
  return tokens.map(token => token.code).join('');
}

function reindent(code, indent = 0) {
  if (typeof code === 'string') {
    if (code.at(0) !== '\n') {
      return ' '.repeat(indent) + code;
    }

    const regex = /^ +/gm;
    const matches = code.matchAll(regex);
    let min = Infinity;

    for (const match of matches) min = Math.min(min, match[0].length);

    const undent = new RegExp(`^ {${min}}`, 'gm');
    const result = code.substring(1).replace(undent, ' '.repeat(indent));
    return result;
  }

  if (code.spans.at(0)?.at(0) !== '\n') return code;
  const regex = /\n */g;
  const matches = code.spans.flatMap(span => [...span.matchAll(regex)]);
  let min = Infinity;

  for (const [match] of matches) min = Math.min(min, match.length - 1);

  const undentRegex = new RegExp(`\n {${min}}`, 'g');
  const spans = code.spans.map(span => span.replace(undentRegex, '\n' + ' '.repeat(indent)));
  spans[0] = spans[0].substring(1);
  let index = 0;
  const nodes = [];

  for (const node of code.nodes) {
    const before = spans[index];
    const preindentRegex = /\n *$/;
    const indentation = before.match(preindentRegex);

    if (indentation != null) {
      spans[index] = spans[index].replace(preindentRegex, '\n');
      nodes.push(reindent(node, indentation[0].length - 1));
    } else {
      nodes.push(node);
    }

    index++;
  }

  return {
    language: code.language,
    spans,
    nodes
  };
}

function integrate(code) {
  if (typeof code === 'string') return code;
  return String.raw({
    raw: code.spans
  }, ...code.nodes.map(n => integrate(n)));
}

function isLevelEquivilant(one, two) {
  if (one === null || two === null) return false;

  if (typeof one === 'string' && typeof two !== 'string' || typeof one !== 'string' && typeof two === 'string') {
    return false;
  }

  if (typeof one === 'string' && typeof two === 'string') {
    return one === two;
  }

  if (typeof one !== 'string' && typeof two !== 'string') {
    if (one.spans.length !== two.spans.length) return false;
    return one.spans.every((span, i) => span === two.spans[i]);
  }

  throw new Error('Could not determine equivilance of nodes');
}

function chars(tokens) {
  return tokens.flatMap(({
    code,
    color
  }) => {
    return code.split('').map(c => ({
      code: c,
      color
    }));
  });
}

function tokens(chars) {
  const result = [];
  let token = null;

  for (const char of chars) {
    if (token == null) {
      token = char;
    } else {
      if (char.code === '\n') {
        result.push(token);
        token = null;
      } else if (token.color === char.color && token.morph === char.morph) {
        token.code += char.code;
      } else {
        result.push(token);
        token = char;
      }
    }
  }

  if (token != null) result.push(token);
  return result;
}

export function diff(start, end, options) {
  start = reindent(start);
  end = reindent(end);
  const startParsed = chars(parse(start, options));
  const endParsed = chars(parse(end, options));
  let index = 0;
  let endex = 0;
  const result = [];

  function recurse(one, two) {
    const progress = (l, r) => {
      if (r == null) r = l;
      const startIndex = index;

      while (index < startIndex + l.length && index < startParsed.length) {
        result.push({
          code: startParsed[index].code,
          color: startParsed[index].color,
          morph: l === r ? 'retain' : 'delete'
        });
        index++;
      }

      const endIndex = endex;

      while (endex < endIndex + r.length && two !== '' && endex < endParsed.length) {
        if (r !== l) {
          result.push({
            code: endParsed[endex].code,
            color: endParsed[endex].color,
            morph: 'create'
          });
        }

        endex++;
      }
    };

    if (isLevelEquivilant(one, two)) {
      if (one == null || two == null) {
        throw new Error('equivilant nodes should not be null');
      }

      if (typeof one === 'string') {
        progress(one);
      } else {
        if (typeof two === 'string') throw new Error();

        for (let n = 0; n < one.nodes.length; n++) {
          progress(one.spans[n]);
          recurse(one.nodes[n], two.nodes[n]);
        }

        progress(one.spans.at(-1) ?? '');
      }
    } else {
      if (typeof one === 'string' && typeof two === 'string') {
        progress(one, two);
      } else if (typeof one === 'string') {
        progress(one, '');
      } else if (typeof two === 'string') {
        progress('', two);
      }

      if (one != null && typeof one !== 'string') {
        for (let n = 0; n < one.nodes.length; n++) {
          progress(one.spans[n], '');
          recurse(one.nodes[n], null);
        }

        progress(one.spans.at(-1) ?? '', '');
      }

      if (two != null && typeof two !== 'string') {
        for (let n = 0; n < two.nodes.length; n++) {
          progress('', two.spans[n]);
          recurse(null, two.nodes[n]);
        }

        progress('', two.spans.at(-1));
      }
    }
  }

  recurse(start, end);
  let [sat, sln, eat, eln] = [0, 0, 0, 0];
  const positioned = result.map(({
    code,
    color,
    morph
  }) => {
    const value = {
      code,
      color,
      morph,
      from: morph === 'create' ? null : [sat, sln],
      to: morph === 'delete' ? null : [eat, eln]
    };

    if (code === '\n') {
      if (morph !== 'create') {
        sat = 0;
        sln++;
      }

      if (morph !== 'delete') {
        eat = 0;
        eln++;
      }
    } else {
      if (morph !== 'create') {
        sat += wcwidth(code);
      }

      if (morph !== 'delete') {
        eat += wcwidth(code);
      }
    }

    return value;
  });
  const tokenized = tokens(positioned);
  return tokenized.filter(({
    code
  }) => {
    return code.length !== 0 && !/^ +$/.test(code);
  });
}

==> code-fns-0.10.1.tgz: package/lib/esm/tags.js <==
import { createStarryNight, all } from '@wooorm/starry-night';
import style from './dark-style';
import wcwidth from 'wcwidth';
const starryNight = createStarryNight(all);
let starryNightCache = null;
export async function ready() {
  starryNightCache = await starryNight;
}
const handler = {
  get(_, language) {
    return (code, ...rest) => {
      return {
        language,
        spans: Array.from(code),
        nodes: rest
      };
    };
  }

};
export const language = new Proxy({}, handler);
export function parse(code, options) {
  const raw = integrate(reindent(code));
  if (starryNightCache == null) throw new Error(`you must await ready()`);
  const sn = starryNightCache;
  const scope = sn.flagToScope(code.language);

  if (typeof scope !== 'string') {
    throw new Error(`language ${code.language} not found`);
  }

  const parsed = sn.highlight(raw, scope);
  return parsed.children.map(child => colorRecurse(child, options?.codeStyle ?? {})).flat().map(({
    color,
    ...rest
  }) => ({ ...rest,
    color: color === '' ? '#c9d1d9' : color
  }));
}
const rules = new Map(Object.entries(style).map(([k, v]) => [k, new Map(Object.entries(v))]));
const styleMap = [[['pl-s'], s => s.stringContent?.text], [['pl-s', 'pl-pds'], s => s.stringPunctuation?.text], [['pl-c'], s => s.comment?.text], [['pl-smi'], s => s.variable?.text], [['pl-v'], s => s.parameter?.text], [['pl-s', 'pl-sr'], s => s.regexp?.content ?? s.regexp?.text], [['pl-s', 'pl-sr', 'pl-pds'], s => s.regexp?.brackets ?? s.regexp?.text], [['pl-s', 'pl-sr', 'pl-k'], s => s.regexp?.flags ?? s.regexp?.text], [['pl-c1'], s => s.literal?.text], [['pl-k'], s => s.keyword?.text], [['pl-en'], s => s.entityName?.text], [['pl-ent'], s => s.entityNameTag?.text]];

function findAssociated(styleMap, target, notFound) {
  let best = notFound;
  let specificity = 0;

  for (const [keys, value] of styleMap) {
    if (keys.length <= specificity) {
      continue;
    }

    let keyIndex = 0;
    let targetIndex = 0;

    while (targetIndex < target.length && keyIndex < keys.length) {
      const key = keys[keyIndex];
      const targetKey = target[targetIndex];

      if (targetKey === key) {
        keyIndex++;
      }

      targetIndex++;
    }

    if (keyIndex === keys.length) {
      // all keys found
      best = value;
      specificity = keys.length;
    }
  }

  return best;
}

export function getColor(classList, codeStyle) {
  const fn = findAssociated(styleMap, classList, () => undefined);
  const result = fn(codeStyle);

  if (result != null) {
    return result;
  }

  const styles = rules.get(`.${classList.at(-1)}`);
  console.assert((styles?.size ?? 0) <= 1, `more styles than just color`);
  return styles?.get('color');
}

function colorRecurse(parsed, codeStyle, classList = []) {
  if (parsed.type === 'text') {
    return [{
      code: parsed.value,
      color: codeStyle.base?.text ?? ''
    }];
  } else if (parsed.type === 'element') {
    const className = parsed.properties?.className ?? [];
    const color = getColor(classList.concat(className), codeStyle);
    const children = parsed.children.map(child => colorRecurse(child, codeStyle, classList.concat(className)));
    const result = [];

    const emit = () => {
      if (temp !== '') {
        if (color != '') {
          result.push({
            code: temp,
            color
          });
        } else {
          result.push({
            code: temp,
            color: ''
          });
        }

        temp = '';
      }
    };

    let temp = '';

    for (const child of children) {
      for (const item of child) {
        if (item.color === '') {
          temp += item.code;
        } else {
          emit();
          result.push(item);
        }
      }

      emit();
    }

    emit();
    return result;
  } else {
    throw new Error();
  }
}

export function toString(tokens) {
  return tokens.map(token => token.code).join('');
}

function reindent(code, indent = 0) {
  if (typeof code === 'string') {
    if (code.at(0) !== '\n') {
      return ' '.repeat(indent) + code;
    }

    const regex = /^ +/gm;
    const matches = code.matchAll(regex);
    let min = Infinity;

    for (const match of matches) min = Math.min(min, match[0].length);

    const undent = new RegExp(`^ {${min}}`, 'gm');
    const result = code.substring(1).replace(undent, ' '.repeat(indent));
    return result;
  }

  if (code.spans.at(0)?.at(0) !== '\n') return code;
  const regex = /\n */g;
  const matches = code.spans.flatMap(span => [...span.matchAll(regex)]);
  let min = Infinity;

  for (const [match] of matches) min = Math.min(min, match.length - 1);

  const undentRegex = new RegExp(`\n {${min}}`, 'g');
  const spans = code.spans.map(span => span.replace(undentRegex, '\n' + ' '.repeat(indent)));
  spans[0] = spans[0].substring(1);
  let index = 0;
  const nodes = [];

  for (const node of code.nodes) {
    const before = spans[index];
    const preindentRegex = /\n *$/;
    const indentation = before.match(preindentRegex);

    if (indentation != null) {
      spans[index] = spans[index].replace(preindentRegex, '\n');
      nodes.push(reindent(node, indentation[0].length - 1));
    } else {
      nodes.push(node);
    }

    index++;
  }

  return {
    language: code.language,
    spans,
    nodes
  };
}

function integrate(code) {
  if (typeof code === 'string') return code;
  return String.raw({
    raw: code.spans
  }, ...code.nodes.map(n => integrate(n)));
}

function isLevelEquivilant(one, two) {
  if (one === null || two === null) return false;

  if (typeof one === 'string' && typeof two !== 'string' || typeof one !== 'string' && typeof two === 'string') {
    return false;
  }

  if (typeof one === 'string' && typeof two === 'string') {
    return one === two;
  }

  if (typeof one !== 'string' && typeof two !== 'string') {
    if (one.spans.length !== two.spans.length) return false;
    return one.spans.every((span, i) => span === two.spans[i]);
  }

  throw new Error('Could not determine equivilance of nodes');
}

function chars(tokens) {
  return tokens.flatMap(({
    code,
    color
  }) => {
    return code.split('').map(c => ({
      code: c,
      color
    }));
  });
}

function tokens(chars) {
  const result = [];
  let token = null;

  for (const char of chars) {
    if (token == null) {
      token = char;
    } else {
      if (char.code === '\n') {
        result.push(token);
        token = null;
      } else if (token.color === char.color && token.morph === char.morph) {
        token.code += char.code;
      } else {
        result.push(token);
        token = char;
      }
    }
  }

  if (token != null) result.push(token);
  return result;
}

export function diff(start, end, options) {
  start = reindent(start);
  end = reindent(end);
  const startParsed = chars(parse(start, options));
  const endParsed = chars(parse(end, options));
  let index = 0;
  let endex = 0;
  const result = [];

  function recurse(one, two) {
    const progress = (l, r) => {
      if (r == null) r = l;
      const startIndex = index;

      while (index < startIndex + l.length && index < startParsed.length) {
        result.push({
          code: startParsed[index].code,
          color: startParsed[index].color,
          morph: l === r ? 'retain' : 'delete'
        });
        index++;
      }

      const endIndex = endex;

      while (endex < endIndex + r.length && two !== '' && endex < endParsed.length) {
        if (r !== l) {
          result.push({
            code: endParsed[endex].code,
            color: endParsed[endex].color,
            morph: 'create'
          });
        }

        endex++;
      }
    };

    if (isLevelEquivilant(one, two)) {
      if (one == null || two == null) {
        throw new Error('equivilant nodes should not be null');
      }

      if (typeof one === 'string') {
        progress(one);
      } else {
        if (typeof two === 'string') throw new Error();

        for (let n = 0; n < one.nodes.length; n++) {
          progress(one.spans[n]);
          recurse(one.nodes[n], two.nodes[n]);
        }

        progress(one.spans.at(-1) ?? '');
      }
    } else {
      if (typeof one === 'string' && typeof two === 'string') {
        progress(one, two);
      } else if (typeof one === 'string') {
        progress(one, '');
      } else if (typeof two === 'string') {
        progress('', two);
      }

      if (one != null && typeof one !== 'string') {
        for (let n = 0; n < one.nodes.length; n++) {
          progress(one.spans[n], '');
          recurse(one.nodes[n], null);
        }

        progress(one.spans.at(-1) ?? '', '');
      }

      if (two != null && typeof two !== 'string') {
        for (let n = 0; n < two.nodes.length; n++) {
          progress('', two.spans[n]);
          recurse(null, two.nodes[n]);
        }

        progress('', two.spans.at(-1));
      }
    }
  }

  recurse(start, end);
  let [sat, sln, eat, eln] = [0, 0, 0, 0];
  const positioned = result.map(({
    code,
    color,
    morph
  }) => {
    const value = {
      code,
      color,
      morph,
      from: morph === 'create' ? null : [sat, sln],
      to: morph === 'delete' ? null : [eat, eln]
    };

    if (code === '\n') {
      if (morph !== 'create') {
        sat = 0;
        sln++;
      }

      if (morph !== 'delete') {
        eat = 0;
        eln++;
      }
    } else {
      if (morph !== 'create') {
        sat += wcwidth(code);
      }

      if (morph !== 'delete') {
        eat += wcwidth(code);
      }
    }

    return value;
  });
  const tokenized = tokens(positioned);
  return tokenized.filter(({
    code
  }) => {
    return code.length !== 0 && !/^ +$/.test(code);
  });
}
